/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "gtkrichtext.h"
#include <gtk/gtkmain.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkselection.h>
#include <math.h>
#include <stdio.h>
#include <libhnj/hsjust.h>

#define SCALE 1
#define HYPHEN_PIXELS 20
#define HYPHEN_PENALTY ( (SCALE) * (SCALE) \
			 * (HYPHEN_PIXELS) * (HYPHEN_PIXELS) )
#define SCROLL_PIXELS            5
#define KEY_SCROLL_PIXELS        15

enum {
  FAILED_INSERT_TEXT,
  FAILED_DELETE_TEXT,
  SELECTION_CHANGED,
  LAST_SIGNAL
};


static GtkEditableClass *parent_class = NULL;
static guint rich_text_signals[LAST_SIGNAL] = { 0 };

/* #define DEBUG 1 */

#ifdef DEBUG
#define D(x) fprintf(stderr, x)
#else
#define D(x)
#endif

static void gtk_rich_text_move_beginning_of_line (GtkRichText *text);
static void gtk_rich_text_move_backward_character (GtkRichText *text);
static void gtk_rich_text_delete_forward_character (GtkRichText *text);
static void gtk_rich_text_move_end_of_line (GtkRichText *text);
static void gtk_rich_text_delete_backward_character (GtkRichText *text);
static void gtk_rich_text_delete_to_line_end (GtkRichText *text);
static void gtk_rich_text_move_forward_character (GtkRichText *text);
static void gtk_rich_text_move_next_line (GtkRichText *text);
static void gtk_rich_text_move_previous_line (GtkRichText *text);
static void gtk_rich_text_delete_line (GtkRichText *text);
static void gtk_rich_text_delete_backward_word (GtkRichText *text);
static void gtk_rich_text_move_backward_word (GtkRichText *text);
static void gtk_rich_text_delete_forward_word (GtkRichText *text);
static void gtk_rich_text_move_forward_word (GtkRichText *text);
static void gtk_rich_text_vertical_move_page (GtkRichText *text,
					      gint distance);
static void gtk_rich_text_move_beginning_of_buffer (GtkRichText *text);
static void gtk_rich_text_move_end_of_buffer (GtkRichText *text);
static void gtk_rich_text_locate_cursor( GtkRichText *text );
static void gtk_rich_text_cursor_coords( GtkRichText *text,
					 gint *x,
					 gint *y );
static void gtk_rich_text_select_word( GtkRichText *text );
static void gtk_rich_text_select_line( GtkRichText *text );
static void gtk_rich_text_set_font( GtkRichText *text,
				    Position start,
				    Position end,
				    GtkRichTextFont *font );

static GtkTextFunction control_keys[26] =
{
  (GtkTextFunction)gtk_rich_text_move_beginning_of_line,    /* a */
  (GtkTextFunction)gtk_rich_text_move_backward_character,   /* b */
  (GtkTextFunction)gtk_editable_copy_clipboard,        /* c */
  (GtkTextFunction)gtk_rich_text_delete_forward_character,  /* d */
  (GtkTextFunction)gtk_rich_text_move_end_of_line,          /* e */
  (GtkTextFunction)gtk_rich_text_move_forward_character,    /* f */
  NULL,                                                /* g */
  (GtkTextFunction)gtk_rich_text_delete_backward_character, /* h */
  NULL,                                                /* i */
  NULL,                                                /* j */
  (GtkTextFunction)gtk_rich_text_delete_to_line_end,        /* k */
  NULL,                                                /* l */
  NULL,                                                /* m */
  (GtkTextFunction)gtk_rich_text_move_next_line,            /* n */
  NULL,                                                /* o */
  (GtkTextFunction)gtk_rich_text_move_previous_line,        /* p */
  NULL,                                                /* q */
  NULL,                                                /* r */
  NULL,                                                /* s */
  NULL,                                                /* t */
  (GtkTextFunction)gtk_rich_text_delete_line,               /* u */
  (GtkTextFunction)gtk_editable_paste_clipboard,       /* v */
  (GtkTextFunction)gtk_rich_text_delete_backward_word,      /* w */
  (GtkTextFunction)gtk_editable_cut_clipboard,         /* x */
  NULL,                                                /* y */
  NULL,                                                /* z */
};

static GtkTextFunction alt_keys[26] =
{
  NULL,                                                /* a */
  (GtkTextFunction)gtk_rich_text_move_backward_word,   /* b */
  NULL,                                                /* c */
  (GtkTextFunction)gtk_rich_text_delete_forward_word,  /* d */
  NULL,                                                /* e */
  (GtkTextFunction)gtk_rich_text_move_forward_word,    /* f */
  NULL,                                           /* g */
  NULL,                                           /* h */
  NULL,                                           /* i */
  NULL,                                           /* j */
  NULL,                                           /* k */
  NULL,                                           /* l */
  NULL,                                           /* m */
  NULL,                                           /* n */
  NULL,                                           /* o */
  NULL,                                           /* p */
  NULL,                                           /* q */
  NULL,                                           /* r */
  NULL,                                           /* s */
  NULL,                                           /* t */
  NULL,                                           /* u */
  NULL,                                           /* v */
  NULL,                                           /* w */
  NULL,                                           /* x */
  NULL,                                           /* y */
  NULL,                                           /* z */
};

static void gtk_rich_text_class_init    (GtkRichTextClass   *klass);
static void gtk_rich_text_init          (GtkRichText        *darea);
static void gtk_rich_text_realize       (GtkWidget          *widget);
static void gtk_rich_text_size_allocate (GtkWidget          *widget,
					 GtkAllocation      *allocation);
static void gtk_rich_text_send_configure (GtkRichText       *darea);
static gint gtk_rich_text_expose        (GtkWidget          *widget,
					 GdkEventExpose     *event);
static gint gtk_rich_text_button_press  (GtkWidget          *widget,
					 GdkEventButton     *event);
static gint gtk_rich_text_button_release(GtkWidget          *widget,
					 GdkEventButton     *event);
static gint gtk_rich_text_motion_notify (GtkWidget         *widget,
					 GdkEventMotion    *event);
/*
static gint gtk_rich_text_button_release(GtkWidget          *widget,
					 GdkEventButton     *event);
*/
static gint  gtk_rich_text_key_press         (GtkWidget         *widget,
                                         GdkEventKey       *event);
/*
static gint  gtk_rich_text_focus_in          (GtkWidget         *widget,
                                         GdkEventFocus     *event);
static gint  gtk_rich_text_focus_out         (GtkWidget         *widget,
                                         GdkEventFocus     *event);
*/
static void gtk_rich_text_draw          (GtkWidget          *widget,
					 GdkRectangle       *area);
static void expose_text                 (GtkRichText        *text, 
					 GdkRectangle       *area, 
					 gboolean            cursor);
static void calculate_selection         (GtkRichText        *text);
static void gtk_rich_text_adjustment    (GtkAdjustment      *adjustment,
					 GtkRichText        *text);
static void gtk_rich_text_disconnect    (GtkAdjustment      *adjustment,
					 GtkRichText        *text);
static void gtk_rich_text_destroy       (GtkObject          *object);
static void adjust_adjs                 (GtkRichText        *text);
static Position gtk_rich_text_find_point(GtkRichText        *text,
					 gint                x, 
					 gint                y);

static gchar *gtk_rich_text_get_chars  (GtkEditable       *editable,
					gint               start,
					gint               end);

static Position decrement_position(Position position,
				   gint distance);
static Position increment_position(Position position,
				   gint distance);
static Position vertical_move(GtkRichText *text,
			      Position position,
			      gint distance);

static Position gtk_rich_text_delete_internal(GtkRichText *text,
					      Position start,
					      Position end,
					      Position *maintain,
					      gint poscount);
static Position gtk_rich_text_insert_internal( GtkRichText *text,
					       Position position,
					       gchar *keyvals,
					       gint length,
					       Position *maintain,
					       gint poscount );

static void gtk_rich_text_set_position( GtkEditable *editable,
					int position );
static Position gtk_rich_text_position( GtkRichText *text,
					int position );
static void gtk_rich_text_insert_as_editable (GtkEditable   *editable,
					      const gchar   *new_text,
					      gint           new_text_length,
					      gint           *position);
static void gtk_rich_text_set_selection (GtkEditable        *editable,
					 gint                start,
					 gint                end);

static void gtk_rich_text_delete_as_editable (GtkEditable        *editable,
					      gint                start,
					      gint                end);
static void gtk_rich_text_update_text (GtkEditable        *editable,
				       gint                startpos,
				       gint                endpos);
static void gtk_rich_text_set_ideal( GtkRichText *text );
static gint motion_timeout( gpointer data );
static void gtk_rich_text_queue_redraw( GtkRichText *text );
static gint redraw_timeout( gpointer data );
static void gtk_rich_text_insert_through_editable (GtkRichText *text,
						   gchar *vals,
						   gint length);
void       gtk_rich_text_set_editable_as_editable( GtkEditable   *editable,
						   gboolean is_editable );

static void
gtk_rich_text_real_set_position (GtkRichText *text,
				 Position pos)
{
  GtkEditable *editable = GTK_EDITABLE( text );
  editable->current_pos = gtk_rich_text_position_to_int( text, pos );
  gtk_editable_select_region( editable,
			      editable->current_pos, editable->current_pos );
  text->selection_end = text->cursor = pos;
  text->selection_active = FALSE;
  text->cursor_first = TRUE;
  if ( text->cursor_font != NULL )
    gtk_rich_text_font_unref( text->cursor_font );
  text->cursor_font = NULL;  
  gtk_rich_text_locate_cursor( text );  
  gtk_rich_text_queue_redraw( text );
}

static void
gtk_rich_text_real_set_position_with_int (GtkRichText *text,
					  Position pos,
					  gint with_int)
{
  GtkEditable *editable = GTK_EDITABLE( text );
  editable->current_pos = with_int;
  gtk_editable_select_region( editable,
			      editable->current_pos, editable->current_pos );
  text->selection_end = text->cursor = pos;
  text->selection_active = FALSE;
  text->cursor_first = TRUE;
  if ( text->cursor_font != NULL )
    gtk_rich_text_font_unref( text->cursor_font );
  text->cursor_font = NULL;  
  gtk_rich_text_locate_cursor( text );  
  gtk_rich_text_queue_redraw( text );
}

guint
gtk_rich_text_get_type (void)
{
  static guint rich_text_type = 0;

  if (!rich_text_type)
    {
      GtkTypeInfo rich_text_info =
      {
	"GtkRichText",
	sizeof (GtkRichText),
	sizeof (GtkRichTextClass),
	(GtkClassInitFunc) gtk_rich_text_class_init,
	(GtkObjectInitFunc) gtk_rich_text_init,
	/* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      rich_text_type = gtk_type_unique (GTK_TYPE_EDITABLE, &rich_text_info);
    }

  return rich_text_type;
}

static void
gtk_rich_text_class_init (GtkRichTextClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkEditableClass *editable_class;

  gtk_word_init();

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;
  editable_class = (GtkEditableClass*) class;

  parent_class = gtk_type_class (GTK_TYPE_EDITABLE);
  
  class->failed_insert_text = NULL;
  class->failed_delete_text = NULL;
  class->selection_changed = NULL;

  widget_class->realize = gtk_rich_text_realize;
  widget_class->size_allocate = gtk_rich_text_size_allocate;
  widget_class->draw = gtk_rich_text_draw;
  widget_class->expose_event = gtk_rich_text_expose;
  widget_class->button_press_event = gtk_rich_text_button_press;
  widget_class->button_release_event = gtk_rich_text_button_release;
  widget_class->motion_notify_event = gtk_rich_text_motion_notify;
  widget_class->key_press_event = gtk_rich_text_key_press;
  ((GtkObjectClass*) class)->destroy = gtk_rich_text_destroy;

  editable_class->set_editable = gtk_rich_text_set_editable_as_editable;
  editable_class->insert_text = gtk_rich_text_insert_as_editable;
  editable_class->delete_text = gtk_rich_text_delete_as_editable;
      
  /*  editable_class->move_cursor = gtk_rich_text_move_cursor;
      editable_class->move_word = gtk_rich_text_move_word;
      editable_class->move_page = gtk_rich_text_move_page;
      editable_class->move_to_row = gtk_rich_text_move_to_row;
      editable_class->move_to_column = gtk_rich_text_move_to_column;
    
      editable_class->kill_char = gtk_rich_text_kill_char;
      editable_class->kill_word = gtk_rich_text_kill_word;
      editable_class->kill_line = gtk_rich_text_kill_line;
  */
  editable_class->update_text = gtk_rich_text_update_text; 
  editable_class->get_chars = gtk_rich_text_get_chars;
  editable_class->set_selection = gtk_rich_text_set_selection;
  editable_class->set_position = gtk_rich_text_set_position;


  rich_text_signals[FAILED_INSERT_TEXT] =
    gtk_signal_new ("failed_insert_text",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkRichTextClass, failed_insert_text),
		    gtk_marshal_NONE__POINTER_INT_POINTER,
		    GTK_TYPE_NONE,
		    3,
		    GTK_TYPE_STRING,
		    GTK_TYPE_INT,
		    GTK_TYPE_POINTER);

  rich_text_signals[FAILED_DELETE_TEXT] =
    gtk_signal_new ("failed_delete_text",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkRichTextClass, failed_delete_text),
		    gtk_marshal_NONE__INT_INT,
		    GTK_TYPE_NONE,
		    2,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT);

  rich_text_signals[SELECTION_CHANGED] =
    gtk_signal_new ("selection_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkRichTextClass, selection_changed),
		    gtk_marshal_NONE__POINTER_POINTER,
		    GTK_TYPE_NONE,
		    2,
		    GTK_TYPE_POINTER,
		    GTK_TYPE_POINTER);

  gtk_object_class_add_signals (object_class, rich_text_signals, LAST_SIGNAL);
#if 0
  void (* changed)      (GtkEditable    *editable);

  /* Bindings actions */
  void (* activate)        (GtkEditable *editable);
  void (* set_editable)    (GtkEditable *editable,
			    gboolean	 is_editable);
  void (* move_cursor)     (GtkEditable *editable,
			    gint         x,
			    gint         y);
  void (* move_word)       (GtkEditable *editable,
			    gint         n);
  void (* move_page)       (GtkEditable *editable,
			    gint         x,
			    gint         y);
  void (* move_to_row)     (GtkEditable *editable,
			    gint         row);
  void (* move_to_column)  (GtkEditable *editable,
			    gint         row);
  void (* kill_char)       (GtkEditable *editable,
			    gint         direction);
  void (* kill_word)       (GtkEditable *editable,
			    gint         direction);
  void (* kill_line)       (GtkEditable *editable,
			    gint         direction);
  void (* cut_clipboard)   (GtkEditable *editable);
  void (* copy_clipboard)  (GtkEditable *editable);
  void (* paste_clipboard) (GtkEditable *editable);

  /* Virtual functions. get_chars is in paricular not a signal because
   * it returns malloced memory. The others are not signals because
   * they would not be particularly useful as such. (All changes to
   * selection and position do not go through these functions)
   */
  void (* update_text)  (GtkEditable    *editable,
			 gint            start_pos,
			 gint            end_pos);
#endif
}

static void
gtk_rich_text_update_text( GtkEditable *editable, gint startpos, gint endpos )
{
  gtk_rich_text_queue_redraw( GTK_RICH_TEXT( editable ) );
}

static void
gtk_rich_text_init (GtkRichText *darea)
{
}

GtkWidget*
gtk_rich_text_new (GtkAdjustment *hadj,
		   GtkAdjustment *vadj)
{
  GtkWidget *new_widget =
    GTK_WIDGET (gtk_type_new (gtk_rich_text_get_type ()));
  GtkRichText *text = GTK_RICH_TEXT( new_widget );
  Position initpos;
  gint i;

  text->is_editable = TRUE;
  text->line_wrap = TRUE;
  text->word_wrap = TRUE;
  text->frozen = FALSE;

  text->draw_data = NULL;
  text->minimum_width = 0;
  text->width = -1;
  text->hadj = NULL;
  text->vadj = NULL;
  text->paragraphs = g_list_append( NULL, gtk_paragraph_new( text ) );
  gtk_paragraph_set_width( (Paragraph *) text->paragraphs->data, 200 );
  
  text->cursor_font = NULL;

  gtk_rich_text_set_adjustments ( text, hadj, vadj);

  initpos.paragraph = text->paragraphs;
  initpos.paragraph_ptr = gtk_paragraph_start( text->paragraphs->data );
  gtk_rich_text_real_set_position( text, initpos );
  text->point_valid = FALSE;
  text->ideal_valid = FALSE;
  
  gtk_widget_set_events(new_widget,
			gtk_widget_get_events(new_widget) |
			GDK_BUTTON_PRESS_MASK |
			GDK_BUTTON1_MOTION_MASK |
			GDK_BUTTON2_MOTION_MASK |
			GDK_BUTTON3_MOTION_MASK |
			GDK_BUTTON_RELEASE_MASK |
			GDK_KEY_PRESS_MASK);

  GTK_WIDGET_SET_FLAGS( new_widget, GTK_CAN_FOCUS );

  return new_widget;
}

guint
gtk_rich_text_get_length      (GtkRichText       *text)
{
  GList *current_paragraph;
  gint length = 0;

  D( "Entering gtk_rich_text_get_length.\n" );
  for ( current_paragraph = text->paragraphs;
	current_paragraph;
	current_paragraph = g_list_next( current_paragraph ) )
    {
      Paragraph *paragraph = (Paragraph *) current_paragraph->data;
      length += gtk_paragraph_length( paragraph );
      length += 1;
    }
  length -= 1;
  D( "Leaving gtk_rich_text_get_length.\n" );

  return length;
}

static void
gtk_rich_text_set_ideal( GtkRichText *text )
{
  if ( text->ideal_valid == FALSE )
    {
      text->ideal =
	gtk_paragraph_find_ideal( (Paragraph *) text->cursor.paragraph->data,
				  text->cursor.paragraph_ptr );
      text->ideal_valid = TRUE;
    }
}

void
gtk_rich_text_freeze          (GtkRichText       *text)
{
  text->frozen = TRUE;
}

void
gtk_rich_text_thaw            (GtkRichText       *text)
{
  text->frozen = FALSE;
  gtk_rich_text_queue_redraw( text );
  adjust_adjs( text );
}

void
gtk_rich_text_set_point       (GtkRichText       *text,
			       guint              index)
{
  D( "Call to gtk_rich_text_set_point.\n" );
  text->point = gtk_rich_text_position( text, index );
  text->point_valid = TRUE;
}

guint
gtk_rich_text_get_point       (GtkRichText       *text)
{
  if ( text->point_valid )
    return gtk_rich_text_position_to_int( text, text->point );
  else
    return 0;
}

void
gtk_rich_text_set_editable    (GtkRichText       *text,
			       gboolean           editable)
{
  text->is_editable = editable;
}

void
gtk_rich_text_set_editable_as_editable( GtkEditable   *editable,
					gboolean is_editable )
{
  gtk_rich_text_set_editable( GTK_RICH_TEXT( editable ), is_editable );
}

void
gtk_rich_text_set_word_wrap   (GtkRichText       *text,
			       gint               word_wrap)
{
  text->word_wrap = word_wrap;
}

void
gtk_rich_text_set_line_wrap   (GtkRichText       *text,
			       gint               line_wrap)
{
  text->line_wrap = line_wrap;
}

static void
gtk_rich_text_set_position( GtkEditable *editable, int position )
{
  GtkRichText *text = GTK_RICH_TEXT (editable);
  g_return_if_fail (editable != NULL);
  g_return_if_fail (GTK_IS_RICH_TEXT (text));

  D( "Entering gtk_rich_text_set_position.\n" );

  gtk_rich_text_real_set_position_with_int( text,
					    gtk_rich_text_position( text,
								    position ),
					    position );
  text->ideal_valid = FALSE;
  
  D( "Leaving gtk_rich_text_set_position.\n" );
}

static Position
gtk_rich_text_position( GtkRichText *text, int position )
{
  Position rval;
  GList *current_paragraph;

  D( "Entering gtk_rich_text_position.\n" );

  for ( current_paragraph = text->paragraphs;
	current_paragraph;
	current_paragraph = current_paragraph->next )
    {
      Paragraph *paragraph = (Paragraph *) current_paragraph->data;
      if ( position <= gtk_paragraph_length( paragraph ) )
	{
	  rval.paragraph = current_paragraph;
	  rval.paragraph_ptr = gtk_paragraph_position( paragraph, position );
	  break;
	}
      else
	{
	  position -= gtk_paragraph_length( paragraph );
	  position -= 1;
	}
    }

  D( "Leaving gtk_rich_text_position.\n" );

  return rval;
}

gint
gtk_rich_text_position_to_int( GtkRichText *text, Position pos )
{
  GList *current_paragraph;
  gint rval = 0;

  D( "Entering gtk_rich_text_position_to_int.\n" );

  for ( current_paragraph = text->paragraphs;
	current_paragraph;
	current_paragraph = current_paragraph->next )
    {
      Paragraph *paragraph = (Paragraph *) current_paragraph->data;
      if ( pos.paragraph == current_paragraph )
	{
	  rval += gtk_paragraph_position_to_int( paragraph, pos.paragraph_ptr );
	  break;
	}
      else
	rval += gtk_paragraph_length( paragraph ) + 1;
    }

  D( "Leaving gtk_rich_text_position_to_int.\n" );

  return rval;
}

void
gtk_rich_text_set_adjustments (GtkRichText   *text,
			       GtkAdjustment *hadj,
			       GtkAdjustment *vadj)
{
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_RICH_TEXT (text));

  D( "Entering gtk_rich_text_set_adjustments.\n" );
  
  if (text->hadj && (text->hadj != hadj))
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (text->hadj), text);
      gtk_object_unref (GTK_OBJECT (text->hadj));
    }
  
  if (text->vadj && (text->vadj != vadj))
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (text->vadj), text);
      gtk_object_unref (GTK_OBJECT (text->vadj));
    }
  
  if (!hadj)
    hadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
  
  if (!vadj)
    vadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
  
  if (text->hadj != hadj)
    {
      text->hadj = hadj;
      gtk_object_ref (GTK_OBJECT (text->hadj));
      gtk_object_sink (GTK_OBJECT (text->hadj));
      
      gtk_signal_connect (GTK_OBJECT (text->hadj), "changed",
			  (GtkSignalFunc) gtk_rich_text_adjustment,
			  text);
      gtk_signal_connect (GTK_OBJECT (text->hadj), "value_changed",
			  (GtkSignalFunc) gtk_rich_text_adjustment,
			  text);
      gtk_signal_connect (GTK_OBJECT (text->hadj), "disconnect",
			  (GtkSignalFunc) gtk_rich_text_disconnect,
			  text);
    }
  
  if (text->vadj != vadj)
    {
      text->vadj = vadj;
      gtk_object_ref (GTK_OBJECT (text->vadj));
      gtk_object_sink (GTK_OBJECT (text->vadj));
      
      gtk_signal_connect (GTK_OBJECT (text->vadj), "changed",
			  (GtkSignalFunc) gtk_rich_text_adjustment,
			  text);
      gtk_signal_connect (GTK_OBJECT (text->vadj), "value_changed",
			  (GtkSignalFunc) gtk_rich_text_adjustment,
			  text);
      gtk_signal_connect (GTK_OBJECT (text->vadj), "disconnect",
			  (GtkSignalFunc) gtk_rich_text_disconnect,
			  text);
    }
  D( "Leaving gtk_rich_text_set_adjustments.\n" );
}

static void
gtk_rich_text_adjustment (GtkAdjustment *adjustment,
			  GtkRichText       *text)
{
  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_RICH_TEXT (text));

  D( "Entering gtk_rich_text_adjustment.\n" );
  
  /* Just ignore it if we haven't been size-allocated yet, or
   * if something weird has happened */
  if ((GTK_WIDGET (text)->allocation.height <= 1) || 
      (GTK_WIDGET (text)->allocation.width <= 1))
    return;
  
  if (adjustment == text->hadj)
    {
      text->xoff = ((gint)adjustment->value);
    }
  else
    {
      text->yoff = ((gint)adjustment->value);
    }
  
  gtk_rich_text_queue_redraw(text);

  D( "Leaving gtk_rich_text_adjustment.\n" );
  
}

static void
gtk_rich_text_disconnect (GtkAdjustment *adjustment,
			  GtkRichText       *text)
{
  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));
  g_return_if_fail (text != NULL);
  g_return_if_fail (GTK_IS_RICH_TEXT (text));

  D( "Entering gtk_rich_text_disconnect.\n" );
  
  if (adjustment == text->hadj)
    text->hadj = NULL;
  if (adjustment == text->vadj)
    text->vadj = NULL;

  D( "Leaving gtk_rich_text_disconnect.\n" );
}

void
gtk_rich_text_set_minimum_width (GtkRichText *text,
				 gint         min_width)
{
  GList *list;
  Paragraph *par;

  D( "Entering gtk_rich_text_set_minimum_width.\n" );

  text->minimum_width = min_width;
  for ( list = text->paragraphs; list != NULL; list = list->next )
    {
      par = (Paragraph *) list->data;
      if ( text->width == -1 )
	gtk_paragraph_set_width( par,
				 MAX
				 ( GTK_WIDGET(text)->allocation.width * SCALE,
				   text->minimum_width ) );
      else
	gtk_paragraph_set_width( par, text->width );
    }
  adjust_adjs(text);
  gtk_rich_text_queue_redraw(text);

  D( "Leaving gtk_rich_text_set_minimum_width.\n" );
}

void
gtk_rich_text_set_width (GtkRichText *text,
			 gint         width)
{
  GList *list;
  Paragraph *par;

  D( "Entering gtk_rich_text_set_width.\n" );

  if ( width != text->width )
    {
      text->width = width;
      for ( list = text->paragraphs; list != NULL; list = list->next )
	{
	  par = (Paragraph *) list->data;
	  if ( text->width == -1 )
	    gtk_paragraph_set_width( par,
				     MAX
				     (GTK_WIDGET(text)->allocation.width*SCALE,
				      text->minimum_width ) );
	  else
	    gtk_paragraph_set_width( par, text->width );
	}
      adjust_adjs(text);
      gtk_rich_text_queue_redraw(text);
    }
  
  D( "Leaving gtk_rich_text_set_width.\n" );
}

gint
gtk_rich_text_get_width (GtkRichText *text)
{
  return text->width;
}

gint
gtk_rich_text_forward_delete(GtkRichText *text, guint nchars)
{
  Position end = increment_position( text->point, nchars );
  Position poss[2];

  D( "Entering gtk_rich_text_forward_delete.\n" );

  if ( ! text->point_valid )
    {
      text->point = gtk_rich_text_position( text, 0 );
      text->point_valid = TRUE;
    }

  poss[0] = text->cursor;
  poss[1] = text->selection_end;
   
  text->point = gtk_rich_text_delete_internal( text, text->point, end,
					       poss, 2 );
  text->ideal_valid = FALSE;
 
  text->cursor = poss[0];
  text->selection_end = poss[1];
  GTK_EDITABLE( text )->current_pos =
    gtk_rich_text_position_to_int( text, text->cursor );
   
  if ( ! text->frozen )
    {
      adjust_adjs( text );
      gtk_rich_text_locate_cursor( text );
      gtk_rich_text_queue_redraw( text );
    }

  D( "Leaving gtk_rich_text_forward_delete.\n" );
  
  return TRUE;
}

gint
gtk_rich_text_backward_delete(GtkRichText *text, guint nchars )
{
  Position start = decrement_position( text->cursor, nchars );
  
  Position poss[2];
   
  D( "Entering gtk_rich_text_backward_delete.\n" );

  if ( ! text->point_valid )
    {
      text->point = gtk_rich_text_position( text, 0 );
      text->point_valid = TRUE;
    }

  poss[0] = text->cursor;
  poss[1] = text->selection_end;
  
  text->point = gtk_rich_text_delete_internal( text, start, text->cursor,
					       poss, 2 );
  text->ideal_valid = FALSE;
 
  text->cursor = poss[0];
  text->selection_end = poss[1];
  GTK_EDITABLE( text )->current_pos =
    gtk_rich_text_position_to_int( text, text->cursor );
  
  if ( ! text->frozen )
    {
      adjust_adjs( text );
      gtk_rich_text_locate_cursor( text );
      gtk_rich_text_queue_redraw( text );
    }

  D( "Leaving gtk_rich_text_backward_delete.\n" );
  
  return TRUE;
}

static void
gtk_rich_text_realize (GtkWidget *widget)
{
  GtkRichText *text;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_RICH_TEXT (widget));

  D( "Entering gtk_rich_text_realize.\n" );

  text = GTK_RICH_TEXT (widget);
  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
				   &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, text);

  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  gtk_rich_text_send_configure (GTK_RICH_TEXT (widget));

  ( (Paragraph *) text->paragraphs->data)->params.hnjparams.max_neg_space
    = 0 * SCALE;
  ( (Paragraph *) text->paragraphs->data)->params.hnjparams.tab_width
    = 36 * SCALE;
  adjust_adjs( text );

  D( "Leaving gtk_rich_text_realize.\n" );
}

static void
gtk_rich_text_size_allocate (GtkWidget     *widget,
				GtkAllocation *allocation)
{
  GList *list;
  Paragraph *par;
  GtkRichText *text;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_RICH_TEXT (widget));
  g_return_if_fail (allocation != NULL);

  D( "Entering gtk_rich_text_size_allocate.\n" );

  text = GTK_RICH_TEXT( widget );

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

      gtk_rich_text_send_configure (GTK_RICH_TEXT (widget));
    }

  for ( list = text->paragraphs; list != NULL; list = list->next )
    {
      par = (Paragraph *) list->data;
      if ( text->width == -1 )
	gtk_paragraph_set_width( par, MAX( allocation->width * SCALE,
					   text->minimum_width ) );
      else
	gtk_paragraph_set_width( par, text->width );
    }
  adjust_adjs( GTK_RICH_TEXT( widget ) );

  D( "Leaving gtk_rich_text_size_allocate.\n" );
}

static void
gtk_rich_text_send_configure (GtkRichText *darea)
{
  GtkWidget *widget;
  GdkEventConfigure event;

  D( "Entering gtk_rich_text_send_configure.\n" );

  widget = GTK_WIDGET (darea);

  event.type = GDK_CONFIGURE;
  event.window = widget->window;
  event.x = widget->allocation.x;
  event.y = widget->allocation.y;
  event.width = widget->allocation.width;
  event.height = widget->allocation.height;
  
  gtk_widget_event (widget, (GdkEvent*) &event);

  D( "Leaving gtk_rich_text_send_configure.\n" );
}

static void
gtk_rich_text_draw (GtkWidget    *widget,
		    GdkRectangle *area)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_RICH_TEXT (widget));
  g_return_if_fail (area != NULL);

  D( "Entering gtk_rich_text_draw.\n" );
  
  if (GTK_WIDGET_DRAWABLE (widget))
    {
      gtk_rich_text_queue_redraw (GTK_RICH_TEXT (widget));
      gtk_widget_draw_focus (widget);
    }

  D( "Leaving gtk_rich_text_draw.\n" );
}

static gint
gtk_rich_text_expose (GtkWidget    *widget,
		      GdkEventExpose *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_RICH_TEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  D( "Entering gtk_rich_text_expose.\n" );

  /*  if (event->window == GTK_RICH_TEXT (widget)->text_area)
    { 
    RTDEBUG (("in gtk_rich_text_expose (expose)\n"));*/
  gtk_rich_text_queue_redraw (GTK_RICH_TEXT (widget));
      /*    }
	    else if (event->count == 0)
	    {
	    RTDEBUG (("in gtk_rich_text_expose (focus)\n"));
	    gtk_widget_draw_focus (widget);
      }*/

  D( "Leaving gtk_rich_text_expose.\n" );
  
  return FALSE;
}

static void
calculate_selection (GtkRichText *text)
{
  GList *current_paragraph; /* Of type Paragraph *. */

  D( "Entering calculate_selection.\n" );

  if ( GTK_WIDGET_REALIZED( GTK_WIDGET( text ) ) )
    {
      for ( current_paragraph = text->paragraphs;
	    current_paragraph;
	    current_paragraph = g_list_next( current_paragraph ) )
	{
	  Paragraph *paragraph = (Paragraph *) current_paragraph->data;
	  gint jump = ((current_paragraph == text->cursor.paragraph) ? 1 : 0 )
	    + ((current_paragraph == text->selection_end.paragraph) ? 2 : 0 );
	  switch( jump )
	    {
	    case 0:
	      break;
	    case 1:
	      text->cursor_first = TRUE;
	      current_paragraph = NULL;
	      break;
	    case 2:
	      text->cursor_first = FALSE;
	      current_paragraph = NULL;
	      break;
	    case 3:
	      if ( gtk_paragraph_position_equal
		   ( paragraph,
		     text->cursor.paragraph_ptr,
		     text->selection_end.paragraph_ptr ) )
		text->cursor_first = text->selection_active = FALSE;
	      else
		text->cursor_first
		  = gtk_paragraph_a_first( paragraph,
					   text->cursor.paragraph_ptr,
					   text->selection_end.paragraph_ptr );
	      current_paragraph = NULL;
	      break;
	      }
	}
    }

  D( "Leaving calculate_selection.\n" );
}

static void
expose_text (GtkRichText* text, GdkRectangle *area, gboolean cursor)
{
  GdkPixmap *pixmap;
  int y;
  GList *current_paragraph; /* Of type Paragraph *. */
  gboolean in_selection = FALSE;

  D( "Entering expose_text.\n" );

  if ( GTK_WIDGET_REALIZED( GTK_WIDGET( text ) ) )
    {

      GdkGC *gc
	= GTK_WIDGET( text )->style->text_gc[ GTK_WIDGET( text )->state ];
      GdkGC *gcsel
	= GTK_WIDGET( text )->style->dark_gc[ GTK_WIDGET( text )->state ];
      pixmap = gdk_pixmap_new( GTK_WIDGET( text )->window,
			       area->width, area->height,
			       gtk_widget_get_visual
			       ( GTK_WIDGET( text ) )->depth );
  
      gdk_draw_rectangle (pixmap,
			  GTK_WIDGET(text)->style->base_gc
			  [GTK_WIDGET(text)->state],
			  TRUE,
			  0, 0,
			  area->width, area->height);

      y = 0;
      
      for ( current_paragraph = text->paragraphs;
	    current_paragraph;
	    current_paragraph = g_list_next( current_paragraph ) )
	{
	  Paragraph *paragraph = (Paragraph *) current_paragraph->data;
	  Position first
	    = text->cursor_first ? text->cursor : text->selection_end;
	  Position second
	    = text->cursor_first ? text->selection_end : text->cursor;
	  gint jump = ((current_paragraph == first.paragraph) ? 1 : 0 )
	    + ((current_paragraph == second.paragraph) ? 2 : 0 );
	  if ( (y >= text->vadj->value + area->y
		- gtk_paragraph_height( paragraph ) ) &&
	       (y <= text->vadj->value + area->y + area->height ) )
	    {
	      switch( jump )
		{
		case 0:
		  gtk_paragraph_draw( paragraph, - text->hadj->value - area->x,
				      y - text->vadj->value - area->y,
				      pixmap,
				      first.paragraph_ptr, FALSE,
				      second.paragraph_ptr, FALSE,
				      in_selection, gc, gcsel );
		  break;
		case 1:
		  gtk_paragraph_draw( paragraph, - text->hadj->value - area->x,
				      y - text->vadj->value - area->y,
				      pixmap,
				      first.paragraph_ptr, cursor,
				      second.paragraph_ptr, FALSE,
				      in_selection, gc, gcsel );
		  in_selection = TRUE;
		  break;
		case 2:
		  gtk_paragraph_draw( paragraph, - text->hadj->value - area->x,
				      y - text->vadj->value - area->y,
				      pixmap,
				      first.paragraph_ptr, FALSE,
				      second.paragraph_ptr, cursor,
				      in_selection, gc, gcsel );
		  in_selection = FALSE;
		  break;
		case 3:
		  gtk_paragraph_draw( paragraph, - text->hadj->value - area->x,
				      y - text->vadj->value - area->y,
				      pixmap,
				      first.paragraph_ptr, cursor,
				      second.paragraph_ptr, cursor,
				      in_selection, gc, gcsel );
		  break;
		}
	    }
	  else
	    {
	      switch( jump )
		{
		case 1:
		  in_selection = TRUE;
		  break;
		case 2:
		  in_selection = FALSE;
		  break;
		deafult:
		  break;
		}
	      if ( y > text->vadj->value + area->y + area->height )
		break;
	    }
	  y += gtk_paragraph_height( paragraph );
	}
      
      /* Copy the pixmap to the window and clean up */
      gdk_draw_pixmap ( GTK_WIDGET( text )->window,
			gc,
			pixmap,
			0, 0,
			area->x,
			area->y,
			area->width, area->height );
      
      gdk_pixmap_unref (pixmap);
    }

  D( "Leaving expose_text.\n" );
}

static void
gtk_rich_text_destroy(GtkObject *object)
{
  GtkRichText *text = GTK_RICH_TEXT( object );

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_RICH_TEXT (object));

  D( "Entering gtk_rich_text_destroy.\n" );

  if (text->hadj)
    {
      gtk_object_unref (GTK_OBJECT (text->hadj));
      text->hadj = NULL;
    }
  if (text->vadj)
    {
      gtk_object_unref (GTK_OBJECT (text->vadj));
      text->vadj = NULL;
    }

  g_list_foreach( text->paragraphs, (GFunc) gtk_paragraph_destroy, NULL );
  g_list_free( text->paragraphs );
  
  GTK_OBJECT_CLASS( parent_class )->destroy (object);

  D( "Leaving gtk_rich_text_destroy.\n" );
}

static void
adjust_adjs( GtkRichText *text )
{
  gint y = 0;
  gint width, height;
  GList *current_paragraph; /* Of type Paragraph *. */

  D( "Entering adjust_adjs.\n" );

  for ( current_paragraph = text->paragraphs;
	current_paragraph;
	current_paragraph = g_list_next( current_paragraph ) )
    {
      y += gtk_paragraph_height( (Paragraph *) current_paragraph->data );
    }
      
  width = GTK_WIDGET( text )->allocation.width;
  height = GTK_WIDGET( text )->allocation.height;
  
  text->vadj->upper          = y;
  text->vadj->step_increment = MIN (text->vadj->upper, (float) SCROLL_PIXELS);
  text->vadj->page_increment = MIN (text->vadj->upper,
				    height - (float) KEY_SCROLL_PIXELS);
  text->vadj->page_size      = MIN (text->vadj->upper, height);
  text->vadj->value          = MIN (text->vadj->value,
				    text->vadj->upper - text->vadj->page_size);
  text->vadj->value          = MAX (text->vadj->value, 0.0);
  
  gtk_signal_emit_by_name (GTK_OBJECT (text->vadj), "changed");

  if ( text->width == -1 )
    text->hadj->upper        = MAX( width * SCALE, text->minimum_width );
  else
    text->hadj->upper        = text->width;
  text->hadj->step_increment = MIN (text->hadj->upper, (float) SCROLL_PIXELS);
  text->hadj->page_increment = MIN (text->hadj->upper,
				    width - (float) KEY_SCROLL_PIXELS);
  text->hadj->page_size      = MIN (text->hadj->upper, width);
  text->hadj->value          = MIN (text->hadj->value,
				    text->hadj->upper - text->hadj->page_size);
  text->hadj->value          = MAX (text->hadj->value, 0.0);
  
  gtk_signal_emit_by_name (GTK_OBJECT (text->hadj), "changed");

  D( "Leaving adjust_adjs.\n" );
}

/* Takes coordinates from the upper left hand corner of the entire
   block of text. */
Position
gtk_rich_text_find_point( GtkRichText *text, int x, int y )
{
  Position newpos;

  D( "Entering gtk_rich_text_find_point.\n" );
  
  if ( GTK_WIDGET_REALIZED( GTK_WIDGET( text ) ) )
    {
      newpos.paragraph = text->paragraphs;
      for ( newpos.paragraph = text->paragraphs;
	    y > gtk_paragraph_height( newpos.paragraph->data )
	      && newpos.paragraph->next;
	    y -= gtk_paragraph_height( newpos.paragraph->data ),
	      newpos.paragraph = newpos.paragraph->next )
	/* Empty for loop. */;
      newpos.paragraph_ptr
	= gtk_paragraph_find_point( (Paragraph *) (newpos.paragraph->data),
				    x, y );
    }
  
  D( "Leaving gtk_rich_text_find_point.\n" );
  
  return newpos;
}

static void
gtk_rich_text_select_word   (GtkRichText         *text)
{
}

static void
gtk_rich_text_select_line   (GtkRichText         *text)
{
}

static gint
gtk_rich_text_button_press  (GtkWidget           *widget,
			     GdkEventButton      *event)
{
  GtkRichText *text;
  GtkEditable *editable;
  gint x;
  gint y;
  static GdkAtom ctext_atom = GDK_NONE;
  
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_RICH_TEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  
  if (ctext_atom == GDK_NONE)
    ctext_atom = gdk_atom_intern ("COMPOUND_TEXT", FALSE);

  text = GTK_RICH_TEXT( widget );
  editable = GTK_EDITABLE( widget );

  x = event->x + text->xoff;
  y = event->y + text->yoff;

  text->last_button = event->button;

  if (!GTK_WIDGET_HAS_FOCUS (widget))
    gtk_widget_grab_focus (widget);

  if (event->button == 1)
    {
      switch (event->type)
	{
	case GDK_BUTTON_PRESS:
	  text->last_count = 1;
	  gtk_rich_text_real_set_position( text,
					   gtk_rich_text_find_point( text,
								     x, y ) );
	  gtk_grab_add (widget);
	  break;
	  
	case GDK_2BUTTON_PRESS:
	  text->last_count = 2;
	  gtk_rich_text_select_word (text);
	  break;
	  
	case GDK_3BUTTON_PRESS:
	  text->last_count = 3;
	  gtk_rich_text_select_line (text);
	  break;
	  
	default:
	  break;
	}
    }
  else if (event->button == 2)
    {
      gtk_rich_text_real_set_position( text, gtk_rich_text_find_point( text,
								       x, y ) );
      gtk_selection_convert (widget, GDK_SELECTION_PRIMARY,
			     ctext_atom, event->time);
    }
  text->ideal_valid = FALSE;
  
  gtk_rich_text_queue_redraw(text);
  
  D( "Leaving gtk_rich_text_button_press.\n" );
  return FALSE;
}

static gint
gtk_rich_text_button_release (GtkWidget      *widget,
			      GdkEventButton *event)
{
  GtkRichText *text;
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_RICH_TEXT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  
  text = GTK_RICH_TEXT (widget);
  
  gtk_grab_remove (widget);
  
  if (text->last_button != event->button)
    return FALSE;
  
  text->last_button = 0;
#if 0
  if (text->timer)
    {
      gtk_timeout_remove (text->timer);
      text->timer = 0;
    }
#endif

  gtk_grab_remove (widget);
  
  return FALSE;
}

static gint
gtk_rich_text_motion_notify  (GtkWidget           *widget,
			      GdkEventMotion      *event)
{
  GtkRichText *text = GTK_RICH_TEXT( widget );
  gint x = event->x + text->xoff;
  gint y = event->y + text->yoff;
  D( "Entering gtk_rich_text_motion_notify.\n" );

  if ( text->motion_callback == 0 )
    {
      text->motion_callback = gtk_idle_add(
					       motion_timeout,
					       (gpointer) text );
    }
  text->queuedx = x;
  text->queuedy = y;
  
  D( "Leaving gtk_rich_text_motion_notify.\n" );
  
  return FALSE;
}

static gint
redraw_timeout( gpointer data )
{
  GtkRichText *text = GTK_RICH_TEXT( data );
  GdkRectangle rect;
  gint x, y;
  gint width, height;

  gdk_window_get_geometry ( GTK_WIDGET( text )->window,
			    &x, &y, &width, &height,
			    NULL );
  
  rect.x = 0;
  rect.y = 0;
  rect.width = width;
  rect.height = height;

  expose_text(text, &rect, TRUE);

  text->redraw_callback = 0;
  return FALSE;
}

static void
gtk_rich_text_queue_redraw  (GtkRichText *text)
{
  if ( text->redraw_callback == 0 )
    {
      text->redraw_callback
	= gtk_timeout_add( 0, redraw_timeout, (gpointer) text );
    }
}

static gint
motion_timeout( gpointer data )
{
  GtkRichText *text = GTK_RICH_TEXT( data );
  GtkEditable *editable = GTK_EDITABLE( text );
  gboolean cursor_first_before = text->cursor_first;

  text->cursor
    = gtk_rich_text_find_point( text, text->queuedx, text->queuedy );
  editable->current_pos = gtk_rich_text_position_to_int( text, text->cursor );
  text->selection_active = TRUE;
  
  calculate_selection( text );
  
  if ( cursor_first_before )
    {
      if ( text->cursor_first )
	{	
	  editable->selection_start_pos =
	    editable->current_pos;
	}
      else
	{
	  editable->selection_start_pos =
	    editable->selection_end_pos;
	  editable->selection_end_pos =
	    editable->current_pos;
	}
    }
  else
    {
      if ( text->cursor_first )
	{
	  editable->selection_end_pos =
	    editable->selection_start_pos;
	  editable->selection_start_pos =
	    editable->current_pos;
	}
      else
	{
	  editable->selection_end_pos =
	    editable->current_pos;
	}
    }
  gtk_rich_text_queue_redraw(text);
  text->motion_callback = 0;
  gtk_signal_emit( GTK_OBJECT (text),
		   rich_text_signals[SELECTION_CHANGED],
		   text->cursor_first ? text->cursor : text->selection_end,
		   text->cursor_first ? text->selection_end : text->cursor );
  

  return FALSE;
}

static void
gtk_rich_text_insert_through_editable (GtkRichText *text,
				       gchar *vals,
				       gint length)
{
  GtkEditable *editable = GTK_EDITABLE( text );
  gint position = editable->current_pos;
  if( text->is_editable )
    {
      gtk_editable_insert_text( editable, vals, length, &position );
    }
  else
    {
      gtk_signal_emit (GTK_OBJECT (text),
		       rich_text_signals[FAILED_INSERT_TEXT],
		       vals, length, &position);
    }
}

static void
gtk_rich_text_delete_through_editable (GtkRichText *text,
				       Position startpos,
				       Position endpos)
{
  gint start = gtk_rich_text_position_to_int( text, startpos );
  gint end = gtk_rich_text_position_to_int( text, endpos );
  if( text->is_editable )
    {
      GtkEditable *editable = GTK_EDITABLE( text );
      gtk_editable_delete_text( editable, start, end );
      text->ideal_valid = FALSE;
    }
  else
    {
      gtk_signal_emit (GTK_OBJECT (text),
		       rich_text_signals[FAILED_DELETE_TEXT], start, end);
    }
}

static gint
gtk_rich_text_key_press (GtkWidget            *widget,
			 GdkEventKey          *event)
{
  GtkRichText *text = GTK_RICH_TEXT(widget);
  GtkEditable *editable = GTK_EDITABLE( widget );
  gint key = event->keyval;
  gboolean return_val = FALSE;
  gint extend_selection;
  gint selected_before;
  gint extend_start;
  gchar value = key;
  
  extend_selection = event->state & GDK_SHIFT_MASK;
  extend_start = FALSE;

  selected_before =
    ! ( text->cursor.paragraph == text->selection_end.paragraph &&
	gtk_paragraph_position_equal
	( (Paragraph *) text->cursor.paragraph->data,
	  text->cursor.paragraph_ptr,
	  text->selection_end.paragraph_ptr )
	);
  

  if ( selected_before )
    {
      switch ( key )
	{
	case GDK_Left:
	case GDK_KP_Left:
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_move_backward_word (text);
	  else
	    gtk_rich_text_move_backward_character( text );
	  return_val = TRUE;
	  break;
	case GDK_Right:
	case GDK_KP_Right:
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_move_forward_word (text);
	  else
	    gtk_rich_text_move_forward_character( text );
	  return_val = TRUE;
	  break;
	case GDK_Up:
	case GDK_KP_Up:
	  if ( event->state & GDK_CONTROL_MASK )
	    gtk_rich_text_vertical_move_page( text, -1 );
	  else
	    gtk_rich_text_move_previous_line( text );
	  return_val = TRUE;
	  break;
	case GDK_Down:
	case GDK_KP_Down:
	  if ( event->state & GDK_CONTROL_MASK )
	    gtk_rich_text_vertical_move_page( text, 1 );
	  else
	    gtk_rich_text_move_next_line( text );
	  return_val = TRUE;
	  break;

	case GDK_Home:
	case GDK_KP_Home:
	  return_val = TRUE;
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_move_beginning_of_buffer( text );
	  else
	    gtk_rich_text_move_beginning_of_line( text );
	  break;

	case GDK_End:
	case GDK_KP_End:
	  return_val = TRUE;
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_move_end_of_buffer( text );
	  else
	    gtk_rich_text_move_end_of_line( text );
	  break;

	case GDK_Page_Down:
	case GDK_KP_Page_Down:
	  return_val = TRUE;
	  gtk_rich_text_vertical_move_page( text, 1 );
	  break;
      
	case GDK_Page_Up:
	case GDK_KP_Page_Up:
	  return_val = TRUE;
	  gtk_rich_text_vertical_move_page( text, -1 );
	  break;
	  
	case GDK_BackSpace:
	case GDK_Delete:
	case GDK_KP_Delete:
	  
	case GDK_exclam:
	case GDK_quotedbl:
	case GDK_numbersign:
	case GDK_dollar:
	case GDK_percent:
	case GDK_ampersand:
	case GDK_apostrophe:
	  /*    case GDK_quoteright:*/
	case GDK_parenleft:
	case GDK_parenright:
	case GDK_asterisk:
	case GDK_plus:
	case GDK_comma:
	case GDK_minus:
	case GDK_period:
	case GDK_slash:
	  
	case GDK_colon:
	case GDK_semicolon:
	case GDK_less:
	case GDK_equal:
	case GDK_greater:
	case GDK_question:
	case GDK_at:
	  
	case GDK_bracketleft:
	case GDK_backslash:
	case GDK_bracketright:
	case GDK_asciicircum:
	case GDK_underscore:
	case GDK_grave:
	  /*    case GDK_quoteleft: */
	  
	case GDK_braceleft:
	case GDK_bar:
	case GDK_braceright:
	case GDK_asciitilde:
	  
	case GDK_space:
	case GDK_KP_Space:

	case GDK_Tab:
	  
	case GDK_KP_Add:
	case GDK_KP_Subtract:
	case GDK_KP_Divide:
	case GDK_KP_Decimal:
	case GDK_KP_Multiply:
	  
	  return_val = TRUE;
	  gtk_editable_delete_selection( editable );
	  break;
	  
	default:
	  if (!((event->state & GDK_CONTROL_MASK)
		|| (event->state & GDK_MOD1_MASK)))
	      {
		if(( key >= GDK_a && key <= GDK_z ) ||
		   ( key >= GDK_A && key <= GDK_Z ) ||
		   ( key >= GDK_0 && key <= GDK_9 ) ||
		   ( key >= GDK_KP_0 && key <= GDK_KP_9 ))
		  {
		    return_val = TRUE;
		    gtk_editable_delete_selection( editable );
		  }
	      }
	  break;
	}
    }
  else
    {
      switch( key )
	{
	case GDK_BackSpace:
	  return_val = TRUE;
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_delete_backward_word( text );
	  else
	    gtk_rich_text_delete_backward_character( text );
	  break;
	case GDK_Delete:
	case GDK_KP_Delete:
	  return_val = TRUE;
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_delete_forward_word (text);
	  else if (event->state & GDK_SHIFT_MASK)
	    {
	      /*	  extend_selection = FALSE; */
	      gtk_editable_cut_clipboard (editable);
	    }
	  else
	    gtk_rich_text_delete_forward_character (text);
	  break;
	case GDK_Left:
	case GDK_KP_Left:
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_move_backward_word (text);
	  else
	    gtk_rich_text_move_backward_character( text );
	  return_val = TRUE;
	  break;
	case GDK_Right:
	case GDK_KP_Right:
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_move_forward_word (text);
	  else
	    gtk_rich_text_move_forward_character( text );
	  return_val = TRUE;
	  break;
	case GDK_Up:
	case GDK_KP_Up:
	  if ( event->state & GDK_CONTROL_MASK )
	    gtk_rich_text_vertical_move_page( text, -1 );
	  else
	    gtk_rich_text_move_previous_line( text );
	  return_val = TRUE;
	  break;
	case GDK_Down:
	case GDK_KP_Down:
	  if ( event->state & GDK_CONTROL_MASK )
	    gtk_rich_text_vertical_move_page( text, 1 );
	  else
	    gtk_rich_text_move_next_line( text );
	  return_val = TRUE;
	  break;

	case GDK_Home:
	case GDK_KP_Home:
	  return_val = TRUE;
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_move_beginning_of_buffer( text );
	  else
	    gtk_rich_text_move_beginning_of_line( text );
	  break;

	case GDK_End:
	case GDK_KP_End:
	  return_val = TRUE;
	  if (event->state & GDK_CONTROL_MASK)
	    gtk_rich_text_move_end_of_buffer( text );
	  else
	    gtk_rich_text_move_end_of_line( text );
	  break;

	case GDK_Page_Down:
	case GDK_KP_Page_Down:
	  return_val = TRUE;
	  gtk_rich_text_vertical_move_page( text, 1 );
	  break;
      
	case GDK_Page_Up:
	case GDK_KP_Page_Up:
	  return_val = TRUE;
	  gtk_rich_text_vertical_move_page( text, -1 );
	  break;
	  
	default:
	  /* Do nothing. */
	  break;
	}
    }
  switch ( key )
    {
    case GDK_exclam:
    case GDK_quotedbl:
    case GDK_numbersign:
    case GDK_dollar:
    case GDK_percent:
    case GDK_ampersand:
    case GDK_apostrophe:
      /*    case GDK_quoteright:*/
    case GDK_parenleft:
    case GDK_parenright:
    case GDK_asterisk:
    case GDK_plus:
    case GDK_comma:
    case GDK_minus:
    case GDK_period:
    case GDK_slash:

    case GDK_colon:
    case GDK_semicolon:
    case GDK_less:
    case GDK_equal:
    case GDK_greater:
    case GDK_question:
    case GDK_at:
      
    case GDK_bracketleft:
    case GDK_backslash:
    case GDK_bracketright:
    case GDK_asciicircum:
    case GDK_underscore:
    case GDK_grave:
      /*    case GDK_quoteleft: */

    case GDK_braceleft:
    case GDK_bar:
    case GDK_braceright:
    case GDK_asciitilde:

    case GDK_space:
    case GDK_KP_Space:
      return_val = TRUE;
      gtk_rich_text_insert_through_editable( text, &value, 1 );
      break;

    case GDK_KP_Add:
      return_val = TRUE;
      value = '+';
      gtk_rich_text_insert_through_editable( text, &value, 1 );
      break;

    case GDK_KP_Subtract:
      return_val = TRUE;
      value = '-';
      gtk_rich_text_insert_through_editable( text, &value, 1 );
      break;

    case GDK_KP_Divide:
      return_val = TRUE;
      value = '/';
      gtk_rich_text_insert_through_editable( text, &value, 1 );
      break;

    case GDK_KP_Decimal:
      return_val = TRUE;
      value = '.';
      gtk_rich_text_insert_through_editable( text, &value, 1 );
      break;

    case GDK_KP_Multiply:
      return_val = TRUE;
      value = '*';
      gtk_rich_text_insert_through_editable( text, &value, 1 );
      break;
      
    case GDK_Tab:
      return_val = TRUE;
      gtk_rich_text_insert_through_editable( text, "\t", 1 );
      break;

    case GDK_Return:
    case GDK_KP_Enter:
      if (event->state & GDK_CONTROL_MASK)
	gtk_signal_emit_by_name (GTK_OBJECT (text), "activate");
      else
	{
	  return_val = TRUE;
	  gtk_rich_text_insert_through_editable( text, "\n", 1 );
	}
      break;
      
    default:
      if (event->state & GDK_CONTROL_MASK)
	{
	  if ((key >= 'A') && (key <= 'Z'))
	    key -= 'A' - 'a';
	      
	  if ((key >= 'a') && (key <= 'z') && control_keys[(int) (key - 'a')])
	    {
	      (* control_keys[(int) (key - 'a')]) (editable, event->time);
	      return_val = TRUE;
	    }
	  break;
	}
      else if (event->state & GDK_MOD1_MASK)
	{
	  if ((key >= 'A') && (key <= 'Z'))
	    key -= 'A' - 'a';
	  
	  if ((key >= 'a') && (key <= 'z') && alt_keys[(int) (key - 'a')])
	    {
	      (* alt_keys[(int) (key - 'a')]) (editable, event->time);
	      return_val = TRUE;
	    }
	}
      else
	{
	  if(( key >= GDK_a && key <= GDK_z ) ||
	     ( key >= GDK_A && key <= GDK_Z ) ||
	     ( key >= GDK_0 && key <= GDK_9 ))
	    {
	      return_val = TRUE;
	      gtk_rich_text_insert_through_editable( text, &value, 1 );
	    }
	  else if ( key >= GDK_KP_0 && key <= GDK_KP_9 )
	    {
	      return_val = TRUE;
	      value = key - GDK_KP_0 + GDK_0;
	      gtk_rich_text_insert_through_editable( text, &value, 1 );
	    }
	}
      break;
    }

  if ( ! text->frozen )
    {
      adjust_adjs( text );
      gtk_rich_text_locate_cursor( text );
      gtk_rich_text_queue_redraw( text ); 
   }

  if ( text->cursor.paragraph == text->selection_end.paragraph &&
       gtk_paragraph_position_equal( (Paragraph *)text->cursor.paragraph->data,
				     text->cursor.paragraph_ptr,
				     text->selection_end.paragraph_ptr ) )
    {
      text->selection_active = FALSE;
    }

  D( "Leaving gtk_rich_text_key_press.\n" );
  return return_val;
}

static Position
gtk_rich_text_delete_internal( GtkRichText *text,
			       Position start,
			       Position end,
			       Position *maintain,
			       gint poscount )
{
  gint i;
  gint *oldposs = NULL;
  gint startpos, endpos;
  GList *paragraph; /* Of type Paragraph */
  ParagraphPosition para_pos = start.paragraph_ptr;
  Position final_pos = start;

  if ( poscount > 0 )
    {
      startpos = gtk_rich_text_position_to_int( text, start );
      endpos = gtk_rich_text_position_to_int( text, end );
      oldposs = g_new( gint, poscount );
      for( i = 0; i < poscount; i++ )
	{
	  oldposs[i] = gtk_rich_text_position_to_int( text, maintain[i] );
	  oldposs[i] = MIN( oldposs[i], MAX( oldposs[i] - (endpos - startpos),
					     startpos ) );
	}
    }

  for ( paragraph = start.paragraph; paragraph && paragraph != end.paragraph; )
    {
      final_pos.paragraph = paragraph;
      final_pos.paragraph_ptr
	= gtk_paragraph_delete( (Paragraph *) paragraph->data,
				para_pos,
				gtk_paragraph_end
				( (Paragraph *) paragraph->data ) );
      if ( g_list_next( paragraph ) )
	{
	  GList *templist = g_list_next( paragraph );

	  final_pos.paragraph_ptr
	    = gtk_paragraph_concat
	    ( (Paragraph *) paragraph->data,
	      (Paragraph *) g_list_next( paragraph )->data );
	  text->paragraphs = g_list_remove_link( text->paragraphs, templist );

	  if ( end.paragraph == templist )
	    end.paragraph = paragraph;
	  g_list_free_1( templist );
	  
	  para_pos = final_pos.paragraph_ptr;
	  if ( end.paragraph_ptr.word == final_pos.paragraph_ptr.word )
	    end.paragraph_ptr.word_ptr += final_pos.paragraph_ptr.word_ptr;
	}
      else
	paragraph = g_list_next( paragraph );
    }
  if( paragraph )
    {
      final_pos.paragraph = paragraph;
      final_pos.paragraph_ptr
	= gtk_paragraph_delete( (Paragraph *) paragraph->data,
				para_pos, end.paragraph_ptr );
    }
  if ( poscount > 0 )
    {
      for ( i = 0; i < poscount; i++ )
	{
	  maintain[i] = gtk_rich_text_position( text, oldposs[i] );
	}
    }
  return final_pos;
}

/* Scanning code to be used later.
   Looks for some byte with the value 0x42.
   x = ((int *)p)[0] ^ 0x42424242;
   if ( (x - 0x01010101) & ~x & 0x80808080 )
   {
       some byte has the value 0x42.
   }
*/

static Position
gtk_rich_text_insert_internal( GtkRichText *text,
			       Position start,
			       gchar *vals,
			       gint length,
			       Position *maintain,
			       gint poscount )
{
  gint i;
  gint laststart = 0;
  gint *oldposs = NULL;
  gint startpos;
  
  D( "Entering gtk_rich_text_insert_internal.\n" );

  if ( poscount > 0 )
    {
      startpos = gtk_rich_text_position_to_int( text, start );
      oldposs = g_new( gint, poscount );
      for( i = 0; i < poscount; i++ )
	{
	  oldposs[i] = gtk_rich_text_position_to_int( text, maintain[i] );
	  if (oldposs[i] >= startpos )
	    oldposs[i] += length;
	}
    }
  
  for ( i = 0; i < length; i++ )
    {
      if ( vals[i] == '\n')
	{
	  GList *temp;
	  if ( laststart != i )
	    {
	      if ( text->cursor_font != NULL )
		start.paragraph_ptr =
		  gtk_paragraph_insert_with_data( (Paragraph *)
						  start.paragraph->data,
						  start.paragraph_ptr,
						  vals + laststart,
						  i - laststart,
						  text->cursor_font );
	      else
		start.paragraph_ptr =
		  gtk_paragraph_insert( (Paragraph *)
					start.paragraph->data,
					start.paragraph_ptr,
					vals + laststart,
					i - laststart );
	    }
	  temp = start.paragraph->next;
	  start.paragraph->next = g_new( GList, 1 );
	  start.paragraph->next->data
	    = gtk_paragraph_break( start.paragraph->data, start.paragraph_ptr );
	  start.paragraph->next->prev = start.paragraph;
	  start.paragraph = start.paragraph->next;
	  start.paragraph->next = temp;
	  if ( start.paragraph->next != NULL )
	    start.paragraph->next->prev = start.paragraph;
	  start.paragraph_ptr = gtk_paragraph_start( start.paragraph->data );
	  laststart = i + 1;
	}
    }
  if ( laststart != i )
    {
      if ( text->cursor_font != NULL )
	start.paragraph_ptr =
	  gtk_paragraph_insert_with_data( (Paragraph *) start.paragraph->data,
					  start.paragraph_ptr,
					  vals + laststart,
					  i - laststart,
					  text->cursor_font );
      else
	start.paragraph_ptr =
	  gtk_paragraph_insert( (Paragraph *) start.paragraph->data,
				start.paragraph_ptr,
				vals + laststart,
				i - laststart );
    }
  if ( poscount > 0 )
    {
      for ( i = 0; i < poscount; i++ )
	{
	  maintain[i] = gtk_rich_text_position( text, oldposs[i] );
	}
    }
  D( "Leaving gtk_rich_text_insert_internal.\n" );
  return start;
}

void
gtk_rich_text_insert( GtkRichText       *text,
		      GtkRichTextFont   *font,
		      GdkColor          *fore,
		      GdkColor          *back,
		      const char        *chars,
		      gint               length )
{
  Position poss[2];
  Position temp;
  gint posscount = 0;
  gint possvalid[2] = { 0, 0 };

  D( "Entering gtk_rich_text_insert.\n" );

  if ( ! text->point_valid )
    {
      text->point = gtk_rich_text_position( text, 0 );
      text->point_valid = TRUE;
    }

  if ( text->cursor.paragraph == text->point.paragraph )
    {
      if ( gtk_paragraph_position_equal
	   ( (Paragraph *) text->cursor.paragraph->data,
	     text->cursor.paragraph_ptr,
	     text->point.paragraph_ptr ) )
	possvalid[0] = 2;
      else
	{
	  poss[posscount] = text->cursor;
	  possvalid[0] = 1;
	  posscount ++;
	}
    }
  if ( text->selection_end.paragraph == text->point.paragraph )
    {
      if ( gtk_paragraph_position_equal
	   ( (Paragraph *) text->selection_end.paragraph->data,
	     text->selection_end.paragraph_ptr,
	     text->point.paragraph_ptr ) )
	possvalid[1] = 2;
      else
	{
	  poss[posscount] = text->selection_end;
	  possvalid[1] = TRUE;
	  posscount ++;
	}
    }

  temp = text->point;
  
  if ( length == -1 )
    length = strlen( chars );
  text->point = gtk_rich_text_insert_internal( text, text->point,
					       (char *) chars, length,
					       poss, posscount );

  if ( font != NULL )
    {
      gtk_rich_text_set_font( text, temp, text->point, font );
    }

  if ( possvalid[1] == 1 )
    {
      posscount --;
      text->selection_end = poss[posscount];
    }
  else if ( possvalid[1] == 2 )
    text->selection_end = text->point;
  if ( possvalid[0] == 1 )
    {
      posscount --;
      text->cursor = poss[posscount];
    }
  else if ( possvalid[0] == 2 )
    text->cursor = text->point;
  
  GTK_EDITABLE( text )->current_pos
    = gtk_rich_text_position_to_int( text, text->cursor );
  
  if ( ! text->frozen )
    {
      adjust_adjs( text );
      gtk_rich_text_locate_cursor( text );
      gtk_rich_text_queue_redraw( text );
    }

  D( "Leaving gtk_rich_text_insert.\n" );
}

gboolean
gtk_rich_text_position_equal( GtkRichText *text, Position a, Position b )
{
  return a.paragraph == b.paragraph &&
    gtk_paragraph_position_equal( (Paragraph *) a.paragraph->data,
				  a.paragraph_ptr,
				  b.paragraph_ptr );
}

static Position
decrement_position( Position old_pos, gint distance )
{
  D( "Entering decrement_position.\n" );
  for ( ; distance != 0; distance -- )
    {
      Paragraph *prg = (Paragraph *) old_pos.paragraph->data;

      if (gtk_paragraph_position_equal (prg,
					old_pos.paragraph_ptr,
					gtk_paragraph_start (prg)))
	{
	  if ( old_pos.paragraph && old_pos.paragraph->prev )
	    {
	      old_pos.paragraph = old_pos.paragraph->prev;
	      old_pos.paragraph_ptr =
		gtk_paragraph_end( (Paragraph *) old_pos.paragraph->data );
	    }
	}
      else
	{
	  old_pos.paragraph_ptr =
	    gtk_paragraph_decrement_position (prg, old_pos.paragraph_ptr, 1);
	}
    }

  D( "Leaving decrement_position.\n" );
  return old_pos;
}

static void
gtk_rich_text_set_font( GtkRichText *text,
			Position start,
			Position end,
			GtkRichTextFont *font )
{
  GList *list;
  ParagraphPosition parstart = start.paragraph_ptr;
  if ( font != NULL )
    {
      for ( list = start.paragraph;
	    list != NULL && list != end.paragraph;
	    list = list->next,
	      (list ?
	       parstart = gtk_paragraph_start( list->data ) :
	       parstart ) )
	{
	  gtk_paragraph_set_font( list->data,
				  parstart, gtk_paragraph_end( list->data ),
				  font );
	}
      
      if ( list )
	{
	  gtk_paragraph_set_font( list->data,
				  parstart, end.paragraph_ptr,
				  font );
	}
    }

  if ( ! text->frozen )
    {
      gtk_rich_text_locate_cursor( text );
      gtk_rich_text_queue_redraw(text);
    }
}

static Position
increment_position( Position old_pos, gint distance )
{
  D( "Entering increment_position.\n" );
  for ( ; distance > 0; distance -- )
    {
      Paragraph *prg = (Paragraph *) old_pos.paragraph->data;

      if ( gtk_paragraph_position_equal (prg,
					 old_pos.paragraph_ptr,
					 gtk_paragraph_end(prg)))
	{
	  if ( old_pos.paragraph && old_pos.paragraph->next )
	    {
	      old_pos.paragraph = old_pos.paragraph->next;
	      old_pos.paragraph_ptr =
		gtk_paragraph_start( (Paragraph *) old_pos.paragraph->data );
	    }
	}
      else
	{
	  old_pos.paragraph_ptr =
	    gtk_paragraph_increment_position (prg, old_pos.paragraph_ptr, 1 );
	}
    }

  D( "Leaving increment_position.\n" );
  return old_pos;
}

static Position
vertical_move( GtkRichText *text, Position old_pos, gint distance )
{
  gtk_rich_text_set_ideal( text );
  old_pos.paragraph_ptr =
    gtk_paragraph_vertical_move( (Paragraph *) old_pos.paragraph->data,
				 old_pos.paragraph_ptr,
				 text->ideal, &distance );
  while ( distance > 0 && old_pos.paragraph->next )
    {
      old_pos.paragraph = old_pos.paragraph->next;
      old_pos.paragraph_ptr =
	gtk_paragraph_start( (Paragraph *) old_pos.paragraph->data );
      distance --;
      old_pos.paragraph_ptr =
	gtk_paragraph_vertical_move( (Paragraph *) old_pos.paragraph->data,
				     old_pos.paragraph_ptr,
				     text->ideal, &distance );
    }
  while ( distance < 0 && old_pos.paragraph->prev )
    {
      old_pos.paragraph = old_pos.paragraph->prev;
      old_pos.paragraph_ptr =
	gtk_paragraph_end( (Paragraph *) old_pos.paragraph->data );
      distance ++;
      old_pos.paragraph_ptr =
	gtk_paragraph_vertical_move( (Paragraph *) old_pos.paragraph->data,
				     old_pos.paragraph_ptr,
				     text->ideal, &distance );
    }
  return old_pos;
}

void
gtk_rich_text_set_cursor_font( GtkRichText *text, GtkRichTextFont *font )
{
  if ( text->cursor_font != NULL )
    gtk_rich_text_font_unref( text->cursor_font );
  text->cursor_font = font;
  gtk_rich_text_font_ref( text->cursor_font );
}

static gchar *
gtk_rich_text_get_chars  (GtkEditable       *editable,
			  gint               start,
			  gint               end)
{
  GtkRichText *text = GTK_RICH_TEXT( editable );
  GList *current_paragraph;
  gchar *buffer;
  int position = 0;
  int length = gtk_rich_text_get_length( text );
  if ( end == -1 ) end = length;
  start = MAX( start, 0 );
  end = MIN( end, length );
  buffer = g_malloc0( end - start + 1 );
  
  for ( current_paragraph = text->paragraphs;
	current_paragraph;
	current_paragraph = g_list_next( current_paragraph ) )
    {
      Paragraph *paragraph = (Paragraph *) current_paragraph->data;
      int parend = position + gtk_paragraph_length( paragraph );
      if ( position < end && parend >= start )
	{
	  gint thisstart = MAX( start, position );
	  gint thisend = MIN( end, parend );
	  gtk_paragraph_get_chars( paragraph,
				   buffer + thisstart - start,
				   thisstart - position,
				   thisend - position);
	}
      position = parend;
      if ( position < end && position >= start )
	buffer[position - start] = '\n';
      position += 1;
    }
  buffer[end - start] = 0;
  return buffer;
}

static void
gtk_rich_text_move_beginning_of_line (GtkRichText *text)
{
  text->ideal_valid = TRUE;
  text->ideal = 0;
  gtk_rich_text_real_set_position( text,
				   vertical_move( text, text->cursor, 0 ) );
  text->ideal_valid = FALSE;
}

static void
gtk_rich_text_move_backward_character (GtkRichText *text)
{
  gtk_rich_text_real_set_position( text,
				   decrement_position( text->cursor, 1 ) );
  text->ideal_valid = FALSE;
}

static void
gtk_rich_text_delete_forward_character (GtkRichText *text)
{
  Position end = increment_position( text->cursor, 1 );
  gtk_rich_text_delete_through_editable( text, text->cursor, end );
}

static void
gtk_rich_text_move_end_of_line (GtkRichText *text)
{
  Position pos = text->cursor;
  pos.paragraph_ptr = gtk_paragraph_end_of_line( text->cursor.paragraph->data,
						 text->cursor.paragraph_ptr);
  gtk_rich_text_real_set_position( text, pos );
  text->ideal_valid = FALSE;
}

static void
gtk_rich_text_delete_backward_character (GtkRichText *text)
{
  Position beginning = decrement_position( text->cursor, 1 );
  gtk_rich_text_delete_through_editable( text, beginning, text->cursor);
}

static void
gtk_rich_text_delete_to_line_end (GtkRichText *text)
{
  Position end = text->cursor;
  end.paragraph_ptr = gtk_paragraph_end_of_line( text->cursor.paragraph->data,
						 text->cursor.paragraph_ptr);
  gtk_rich_text_delete_through_editable( text, text->cursor, end );
}

static void
gtk_rich_text_move_forward_character (GtkRichText *text)
{
  gtk_rich_text_real_set_position( text,
				   increment_position( text->cursor, 1 ) );
  text->ideal_valid = FALSE;
}

static void
gtk_rich_text_move_next_line (GtkRichText *text)
{
  gtk_rich_text_real_set_position( text,
				   vertical_move( text, text->cursor, 1 ) );
}

static void
gtk_rich_text_move_previous_line (GtkRichText *text)
{
  gtk_rich_text_real_set_position( text,
				   vertical_move( text, text->cursor, -1 ) );
}

static void
gtk_rich_text_delete_line (GtkRichText *text)
{
}

static void
gtk_rich_text_delete_backward_word (GtkRichText *text)
{
  Position end = text->cursor;
  ParagraphPosition pos = end.paragraph_ptr;
  end.paragraph_ptr =
    gtk_paragraph_move_backward_word( text->cursor.paragraph->data,
				      text->cursor.paragraph_ptr );
  if ( gtk_paragraph_position_equal( (Paragraph *) end.paragraph->data,
				     pos, end.paragraph_ptr )
       && end.paragraph->prev )
    {
      end.paragraph = end.paragraph->prev;
      end.paragraph_ptr = gtk_paragraph_end( end.paragraph->data );
      end.paragraph_ptr =
	gtk_paragraph_move_backward_word( end.paragraph->data,
					  end.paragraph_ptr );
    }
  text->ideal_valid = FALSE;
  gtk_rich_text_delete_through_editable( text, end, text->cursor );
}

static void
gtk_rich_text_move_backward_word (GtkRichText *text)
{
  Position pos = text->cursor;
  Position newpos = pos;
  newpos.paragraph_ptr =
    gtk_paragraph_move_backward_word( newpos.paragraph->data,
				      newpos.paragraph_ptr );
  if ( gtk_paragraph_position_equal( (Paragraph *) newpos.paragraph->data,
				     pos.paragraph_ptr, newpos.paragraph_ptr )
       && newpos.paragraph->prev )
    {
      newpos.paragraph = newpos.paragraph->prev;
      newpos.paragraph_ptr = gtk_paragraph_end( newpos.paragraph->data );
      newpos.paragraph_ptr =
	gtk_paragraph_move_backward_word( newpos.paragraph->data,
					  newpos.paragraph_ptr );
    }
  text->ideal_valid = FALSE;
  gtk_rich_text_real_set_position( text, newpos );
}

static void
gtk_rich_text_delete_forward_word (GtkRichText *text)
{
  Position end = text->cursor;
  ParagraphPosition pos = end.paragraph_ptr;
  end.paragraph_ptr =
    gtk_paragraph_move_forward_word( text->cursor.paragraph->data,
				     text->cursor.paragraph_ptr );
  if ( gtk_paragraph_position_equal( (Paragraph *) end.paragraph->data,
				     pos, end.paragraph_ptr )
       && end.paragraph->next )
    {
      end.paragraph = end.paragraph->next;
      end.paragraph_ptr = gtk_paragraph_start( end.paragraph->data );
      end.paragraph_ptr =
	gtk_paragraph_move_forward_word( end.paragraph->data,
					 end.paragraph_ptr );
    }
  gtk_rich_text_delete_through_editable( text, text->cursor, end );
}

static void
gtk_rich_text_move_forward_word (GtkRichText *text)
{
  Position pos = text->cursor;
  Position newpos = pos;
  newpos.paragraph_ptr =
    gtk_paragraph_move_forward_word( newpos.paragraph->data,
				     newpos.paragraph_ptr );
  if ( gtk_paragraph_position_equal( (Paragraph *) newpos.paragraph->data,
				     pos.paragraph_ptr, newpos.paragraph_ptr )
       && newpos.paragraph->next )
    {
      newpos.paragraph = newpos.paragraph->next;
      newpos.paragraph_ptr = gtk_paragraph_start( newpos.paragraph->data );
      newpos.paragraph_ptr =
	gtk_paragraph_move_forward_word( newpos.paragraph->data,
					 newpos.paragraph_ptr );
    }
  text->ideal_valid = FALSE;
  gtk_rich_text_real_set_position( text, newpos );
}

static void
gtk_rich_text_vertical_move_page (GtkRichText *text, gint distance)
{
  gint x, y;
  gtk_rich_text_cursor_coords( text, NULL, &y );
  gtk_rich_text_set_ideal( text );
  x = text->ideal;
  y += text->vadj->page_increment * distance;
  gtk_rich_text_real_set_position( text,
				   gtk_rich_text_find_point( text, x, y ) );
}

static void gtk_rich_text_move_end_of_buffer (GtkRichText *text)
{
  Position pos;
  pos.paragraph = g_list_last( text->paragraphs );
  pos.paragraph_ptr = gtk_paragraph_end( pos.paragraph->data );
  gtk_rich_text_real_set_position( text, pos );
}

static void gtk_rich_text_move_beginning_of_buffer (GtkRichText *text)
{
  Position pos;
  pos.paragraph = text->paragraphs;
  pos.paragraph_ptr = gtk_paragraph_start( pos.paragraph->data );
  gtk_rich_text_real_set_position( text, pos );
}

static gint
gtk_rich_text_top_of_cursor( GtkRichText *text )
{
  gint top = 0;
  GList *list;
  Paragraph *par;
  for( list = text->paragraphs; list != NULL; list = list->next )
    {
      par = (Paragraph *) list->data;
      if ( list == text->cursor.paragraph )
	{
	  top += gtk_paragraph_top_of_pos( par, text->cursor.paragraph_ptr );
	  break;
	}
      else
	{
	  top += gtk_paragraph_height( par );
	}
    }
  return top;
}

static gint
gtk_rich_text_bottom_of_cursor( GtkRichText *text )
{
  gint top = 0;
  GList *list;
  Paragraph *par;
  for( list = text->paragraphs; list != NULL; list = list->next )
    {
      par = (Paragraph *) list->data;
      if ( list == text->cursor.paragraph )
	{
	  top += gtk_paragraph_bottom_of_pos( par,
					      text->cursor.paragraph_ptr );
	  break;
	}
      else
	{
	  top += gtk_paragraph_height( par );
	}
    }
  return top;
}

static void
gtk_rich_text_locate_cursor( GtkRichText *text )
{
  gint height;

  D( "Entering adjust_adjs.\n" );

  height = GTK_WIDGET( text )->allocation.height;
  
  text->vadj->value = MIN( text->vadj->value,
			   gtk_rich_text_top_of_cursor( text ) );
  text->vadj->value = MAX( text->vadj->value,
			   gtk_rich_text_bottom_of_cursor( text ) - height );
  
  gtk_signal_emit_by_name (GTK_OBJECT (text->vadj), "changed");
}

static void
gtk_rich_text_cursor_coords( GtkRichText *text, gint *x, gint *y )
{
  gint height = 0;
  GList *list;
  Paragraph *par;
  for( list = text->paragraphs; list != NULL; list = list->next )
    {
      par = (Paragraph *) list->data;
      if ( list == text->cursor.paragraph )
	{
	  gtk_paragraph_position_coords( par, text->cursor.paragraph_ptr,
					 x, y );
	  if ( y != NULL )
	    (*y) += height;
	  break;
	}
      else
	{
	  height += gtk_paragraph_height( par );
	}
    }
}

static void
gtk_rich_text_insert_as_editable (GtkEditable       *editable,
				  const gchar       *new_text,
				  gint               new_text_length,
				  gint               *position)
{
  GtkRichText *text = GTK_RICH_TEXT( editable );
  Position poss[3];
  Position insertpoint = gtk_rich_text_position( text, *position );

  poss[0] = text->cursor;
  poss[1] = text->selection_end;
  if ( text->point_valid )
    {
      poss[2] = text->point;
    }
 
  insertpoint =
    gtk_rich_text_insert_internal( text, insertpoint,
				   (gchar *) new_text, new_text_length,
				   poss,
				   text->point_valid ? 3 : 2 );
  
  text->cursor = poss[0];
  text->selection_end = poss[1];
  if ( text->point_valid )
    {
      text->point = poss[2];
    }
  GTK_EDITABLE( text )->current_pos =
    gtk_rich_text_position_to_int( text, text->cursor );

  *position = gtk_rich_text_position_to_int( text, insertpoint );

  if ( ! text->frozen )
    {
      adjust_adjs( text );
      gtk_rich_text_locate_cursor( text );
      gtk_rich_text_queue_redraw( text );
    }
}

static void
gtk_rich_text_delete_as_editable (GtkEditable       *editable,
				  gint               start,
				  gint               end)
{
  GtkRichText *text = GTK_RICH_TEXT( editable );
  Position a = gtk_rich_text_position( text, start );
  Position b = gtk_rich_text_position( text, end );
  Position poss[3];

  poss[0] = text->cursor;
  poss[1] = text->selection_end;
  if ( text->point_valid )
    {
      poss[2] = text->point;
    }
 
  gtk_rich_text_delete_internal( text, a, b, poss,
				 text->point_valid ? 3 : 2 );

  text->cursor = poss[0];
  text->selection_end = poss[1];
  if ( text->point_valid )
    {
      text->point = poss[2];
    }
  editable->current_pos = gtk_rich_text_position_to_int( text, text->cursor );
  if ( text->cursor_first )
    {
      editable->selection_start_pos = editable->current_pos;
      editable->selection_end_pos =
	gtk_rich_text_position_to_int( text, text->selection_end );
    }
  else
    {
      editable->selection_start_pos =
	gtk_rich_text_position_to_int( text, text->selection_end );
      editable->selection_end_pos = editable->current_pos;
    }
}

static void
gtk_rich_text_set_selection (GtkEditable        *editable,
			     gint                start_pos,
			     gint                end_pos)
{
  GtkRichText *text = GTK_RICH_TEXT( editable );
  if( start_pos > end_pos )
    {
      gint temp = start_pos;
      start_pos = end_pos;
      end_pos = temp;
    }
  editable->selection_start_pos = start_pos;
  editable->selection_end_pos = end_pos;
  text->cursor = gtk_rich_text_position( text, start_pos );
  if ( start_pos == end_pos )
    {
      text->selection_end = text->cursor;
      text->selection_active = FALSE;
    }
  else
    {
      text->selection_end = gtk_rich_text_position( text, end_pos );
      text->selection_active = TRUE;
      text->cursor_first = TRUE;
    }
  gtk_rich_text_queue_redraw( text );
}

static void
gtk_rich_text_set_paragraph_params (GtkRichText *text,
				    Position a,
				    Position b,
				    ParagraphParams *params,
				    ParagraphParamsValid *validity)
{
  GList *list;
  for ( list = a.paragraph;
	list != NULL && list != b.paragraph;
	list = list->next )
    {
      gtk_paragraph_set_params( list->data, params, validity );
    }
  if ( list )
    gtk_paragraph_set_params( list->data, params, validity );
  adjust_adjs( text );
  gtk_rich_text_queue_redraw( text );
}

void
gtk_rich_text_set_selection_paragraph_params (GtkRichText *text,
					      ParagraphParams *params,
					      ParagraphParamsValid *validity)
{
  Position first = text->cursor_first ? text->cursor : text->selection_end;
  Position second = text->cursor_first ? text->selection_end : text->cursor;
  gtk_rich_text_set_paragraph_params( text, first, second, params, validity );
}

void
gtk_rich_text_set_location_paragraph_params (GtkRichText *text,
					     gint location,
					     ParagraphParams *params,
					     ParagraphParamsValid *validity)
{
  Position first;
  Position second;
  first = second = gtk_rich_text_position( text, location );
  gtk_rich_text_set_paragraph_params( text, first, second, params, validity );
}

void
gtk_rich_text_set_range_paragraph_params (GtkRichText *text,
					  gint start,
					  gint end,
					  ParagraphParams *params,
					  ParagraphParamsValid *validity)
{
  Position first;
  Position second;
  first = gtk_rich_text_position( text, start );
  second = gtk_rich_text_position( text, end );
  gtk_rich_text_set_paragraph_params( text, first, second, params, validity );
}

void
gtk_rich_text_get_paragraph_params (GtkRichText *text,
				    Position a,
				    Position b,
				    ParagraphParams *params,
				    ParagraphParamsValid *validity)
{
  GList *list;
  list = a.paragraph;
  gtk_paragraph_get_params( list->data, params, validity );
  for ( list = a.paragraph;
	list != NULL && list != b.paragraph;
	list = list->next )
    {
      gtk_paragraph_get_params_match( list->data, params, validity );
    }
  if ( list )
    gtk_paragraph_get_params_match( list->data, params, validity );
}

void
gtk_rich_text_get_selection_paragraph_params (GtkRichText *text,
					     ParagraphParams *params,
					     ParagraphParamsValid *validity)
{
  GList *list;
  Position a = text->cursor_first ? text->cursor : text->selection_end;
  Position b = text->cursor_first ? text->selection_end : text->cursor;
  gtk_rich_text_get_paragraph_params( text, a, b, params, validity );
}

void
gtk_rich_text_set_selection_character_params (GtkRichText *text,
					      CharacterParams *params,
					      CharacterParamsValid *validity)
{
  Position first = text->cursor_first ? text->cursor : text->selection_end;
  Position second = text->cursor_first ? text->selection_end : text->cursor;
  GtkRichTextFont *font = params->font;
  if ( validity->font_valid )
    gtk_rich_text_set_font( text, first, second, font );
}

void
gtk_rich_text_set_location_character_params (GtkRichText *text,
					     gint location,
					     CharacterParams *params,
					     CharacterParamsValid *validity)
{
  Position first;
  Position second;
  GtkRichTextFont *font = params->font;
  if ( validity->font_valid )
    {
      first = second = gtk_rich_text_position( text, location );
      gtk_rich_text_set_font( text, first, second, font );
    }
}

void
gtk_rich_text_set_range_character_params (GtkRichText *text,
					  gint start,
					  gint end,
					  CharacterParams *params,
					  CharacterParamsValid *validity)
{
  Position first;
  Position second;
  GtkRichTextFont *font = params->font;
  if ( validity->font_valid )
    {
      first = gtk_rich_text_position( text, start );
      second = gtk_rich_text_position( text, end );
      gtk_rich_text_set_font( text, first, second, font );
    }
}

void
gtk_rich_text_traverse( GtkRichText *text,
			GtkRichTextTraversalCallbacks *callbacks,
			gpointer data )
{
  GList *current_paragraph;
  for ( current_paragraph = text->paragraphs;
	current_paragraph;
	current_paragraph = g_list_next( current_paragraph ) )
    {
      gtk_paragraph_traverse( (Paragraph *) current_paragraph->data,
			      callbacks, data );
    }
}


void
gtk_rich_text_traverse_partial( GtkRichText *text,
				gint start,
				gint end,
				GtkRichTextTraversalCallbacks *callbacks,
				gpointer data )
{
  GList *current_paragraph;
  int position = 0;
  int length = gtk_rich_text_get_length( text );
  if ( end == -1 ) end = length;
  start = MAX( start, 0 );
  end = MIN( end, length );

  for ( current_paragraph = text->paragraphs;
	current_paragraph && position <= end;
	current_paragraph = g_list_next( current_paragraph ) )
    {
      Paragraph *paragraph = (Paragraph *) current_paragraph->data;
      int parend = position + gtk_paragraph_length( paragraph );
      if ( position <= end && parend >= start )
	{
	  gint thisstart = MAX( start, position );
	  gint thisend = MIN( end, parend );
	  gtk_paragraph_traverse_partial( paragraph,
					  thisstart - position,
					  thisend - position,
					  callbacks, data );
	}
      position = parend;
      position += 1;
    }
}

GtkRichTextFont *
gtk_rich_text_font_new( gchar *desc )
{
  GtkRichTextFont *return_val = g_new( GtkRichTextFont, 1 );
  if ( desc == NULL || *desc == 0 )
    desc = "fixed";
  return_val->font = gdk_font_load( desc );
  if ( return_val->font != NULL )
    {
      gdk_font_ref( return_val->font );
      return_val->desc = g_strdup( desc );
      return_val->ref_count = 0;
    }
  else
    {
      g_free( return_val );
      return_val = NULL;
    }
  return_val->data = NULL;
  return return_val;
}

GtkRichTextFont *
gtk_rich_text_font_adopt( GdkFont *font,
			  gchar *desc,
			  gpointer data )
{
  if ( font != NULL )
    {
      GtkRichTextFont *return_val = g_new( GtkRichTextFont, 1 );
      return_val->font = font;
      gdk_font_ref( return_val->font );
      if ( desc != NULL )
	return_val->desc = g_strdup( desc );
      else
	return_val->desc = g_strdup( "" );
      return_val->ref_count = 0;
      return_val->data = data;
      return return_val;
    }
  else
    {
      return NULL;
    }
}

void
gtk_rich_text_font_ref( GtkRichTextFont *font )
{
  font->ref_count ++;
}

void
gtk_rich_text_font_unref( GtkRichTextFont *font )
{
  font->ref_count --;
  if ( font->ref_count <= 0 )
    {
      gdk_font_unref( font->font );
      g_free( font->desc );
      g_free( font );
    }
}

gint
gtk_rich_text_get_position_line_number( GtkRichText *text )
{
  gint line_number = 0;
  GList *list;
  for ( list = text->paragraphs;
	list != text->cursor.paragraph;
	list = g_list_next( list ) )
    {
      line_number += gtk_paragraph_line_count( (Paragraph *) list->data );
    }
  return line_number
    + gtk_paragraph_get_line_number( (Paragraph *) list->data,
				     text->cursor.paragraph_ptr );
}

void
gtk_rich_text_render( GtkRichText *text,
		      GtkRichTextRenderSystem *system )
{
  GList *current_paragraph; /* Of type Paragraph *. */
  for ( current_paragraph = text->paragraphs;
	current_paragraph;
	current_paragraph = g_list_next( current_paragraph ) )
    {
      Paragraph *paragraph = (Paragraph *) current_paragraph->data;
      gtk_paragraph_render( paragraph,
			    system );
    }
  system->output_page( system );
}
